/*
 * This file is part of SubLink.
 * Copyright (C) 2009  Jeferson Hultmann <hultmann@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

"use strict";

var EXPORTED_SYMBOLS = ["buildResolveMenu"];

Components.utils.import("resource://sublink-migreme-4.1.0/util.js");
Components.utils.import("resource://sublink-migreme-4.1.0/util2.js");


function buildResolveMenu(mainMenu, cmdData, isUrlbar) {
  if (RedirServer.isRedirUri(cmdData.uri) === false) {
    console.trace();
    //throw new Error("buildResolveMenu 0 " + cmdData.uri.spec);
  }

  var doc = mainMenu.ownerDocument;
  var menuParent = doc.getElementById("sublink-migreme-menu-resolve");
  if (menuParent) {
    console.trace();
    throw new Error("buildResolveMenu 1 " + menuParent.tagName);
  }

  var sepItem = mainMenu.insertBefore(doc.createElement("menuseparator"), mainMenu.firstChild);

  menuParent = mainMenu.insertBefore(doc.createElement("menu"), mainMenu.firstChild);
  menuParent.setAttribute("id", "sublink-migreme-menu-resolve");
  menuParent.setAttribute("label", cmdData.uri.spec);
  menuParent.classList.add("sublink-migreme-node");
  menuParent.classList.add("menu-iconic");

  sepItem.classList.add("sublink-migreme-node");


  if (cmdData.typed) {
    var copy = isUrlbar || cmdData.node.readOnly;
    menuParent.value = copy ? "popup-input-readonly" : "popup-input-default";
  } else {
    menuParent.value = "popup-link";
  }


  var menuNew = menuParent.appendChild(doc.createElement("menupopup"));
  menuNew.setAttribute("id", "sublink-migreme-menu-resolve-tmp1");



  var pathMenuItem = menuNew.appendChild(doc.createElement("menuitem"));
  pathMenuItem.setAttribute("id", "sublink-migreme-menu-resolve-tmp2");
  pathMenuItem.setAttribute("label", util.getText("context-resolve.working.label"));
  pathMenuItem.setAttribute("disabled", "true");

  menuNew.appendChild(doc.createElement("menuseparator"));



  var copyMenuItem = menuNew.appendChild(doc.createElement("menuitem"));
  copyMenuItem.setAttribute("id", "sublink-migreme-menu-resolve-tmp3");
  copyMenuItem.setAttribute("disabled", "true");
  configMenuItemAction(copyMenuItem, menuParent.value);


  var convertMenuItem = menuNew.appendChild(doc.createElement("menuitem"));
  convertMenuItem.setAttribute("id", "sublink-migreme-menu-resolve-tmp4");
  convertMenuItem.setAttribute("disabled", "true");
  convertMenuItem.setAttribute("label", "convertMenuItem");


  var libraryItem = menuNew.appendChild(doc.createElement("menuitem"));
  libraryItem.setAttribute("id", "sublink-migreme-menu-resolve-tmp5");
  libraryItem.setAttribute("disabled", "true");
  libraryItem.setAttribute("label", "libraryItem");

  menuNew.appendChild(doc.createElement("menuseparator"));




  if (cmdData.uri.host === "migre.me") {//basedomain
    configMenuItemRemember(libraryItem);
    convertMenuItem.hidden = true;

  } else {
    configMenuItemConvert(convertMenuItem, menuParent.value);
    libraryItem.hidden = true;
  }


  var redirPath = redirCache().get(cmdData.uri.spec);
  if (redirPath !== null) {
    console.log("getUserData found", redirPath);
    finishResolvePopup(mainMenu.ownerDocument, cmdData, redirPath);
    var menuNew = doc.getElementById("sublink-migreme-menu-resolve-tmp1");
    updateRedirPath(menuNew, redirPath);
    return;
  }


  // TODO long url pode estar na lib, mas ainda eh necessario mapear o redir
  //if (redirStorage.loadShortUrls(longUri).hasOwnProperty(cmdData.uri.spec)) {
  var online = util.isOffline() === false;
  if (online) {
    menuParent.setAttribute("image", "chrome://browser/skin/tabbrowser/loading.png");
    findRedirCore(mainMenu, cmdData);
  }

}


/*
<INPUT EDIT>    LONG_URL
  *  Compactar com tinyurl: "http://longurl.com/..."

<INPUT READONLY>LONG_URL
  *  Copiar como tinyurl.com "http://longurl.com/..."


<INPUT EDIT>    SHORT_URL
   ! http://longurl.com/...>  Descompactar           [input=>replace]
                              Converter para tinyurl [input=>convert_replace]  (p/ SHORT_URL!=tinyurl)
<INPUT READONLY>SHORT_URL
   ! http://longurl.com/...>  Copiar                 [input=>copy]
                              Copiar como tinyurl    [input=>convert_copy]      (p/ SHORT_URL!=tinyurl)

<A>             SHORT_URL
  ! http://longurl.com/...>   Copiar                 [link=>copy]
                              Copiar como tinyurl    [link=>convert_copy]       (p/ SHORT_URL!=tinyurl)
*/


function findRedirCore(appMenu, cmdData) {
  var myRedirPath = [{ // BUG race condition
    delta: -1,
    responseStatusCode: -1,
    longUri: cmdData.uri
  }];


  var doc = appMenu.ownerDocument;
  var menuNew = doc.getElementById("sublink-migreme-menu-resolve-tmp1");
  updateRedirPath(menuNew, myRedirPath);


  Components.utils.import("resource://sublink-migreme-4.1.0/resolve-redir.js");
  var req = new ResolveRedir(); // TODO redir path usar apenas resolveredir; usar API simultaneamente para outros infos

  req.onLoad = function(info) { // redir not found
    myRedirPath.push(info);
    updateRedirPath(menuNew, myRedirPath);
    finishResolvePopup(doc, cmdData, myRedirPath);
  };

  req.onRedirFound = function(info, qty) {
    myRedirPath.push(info);
    updateRedirPath(menuNew, myRedirPath);

    Components.utils.import("resource://sublink-migreme-4.1.0/util2.js");
    if (RedirServer.isRedirUri(info.longUri)) {
      if (qty < 20) {
        req.findRedir(info.longUri);
      }

    } else {
      // BUG http://tinyurl.com/7ggkqey ==> magnet: abre app
      finishResolvePopup(doc, cmdData, myRedirPath);
      redirCache().set(cmdData.uri.spec, myRedirPath);
    }
  };

  req.onError = function(errData) {
    console.log("FAIL", errData.toSource());
    if (appMenu.state !== "open") {
      return;
      finishResolvePopup(doc, cmdData, myRedirPath);
    }

    function cmdShow(evt) {
      Components.utils.import("resource://sublink-migreme-4.1.0/err.js");
      notificationError(errData, doc.defaultView.getBrowser().selectedBrowser);
    }

    var menu = doc.getElementById("sublink-migreme-menu-resolve");
    menu.setAttribute("image", "chrome://global/skin/icons/error-16.png");
    menu.removeAttribute("disabled");
    menu.addEventListener("popuphidden", onPopupHidden, false);

    var item = menu.appendChild(doc.createElement("menupopup"))
                   .appendChild(doc.createElement("menuitem"));
    item.setAttribute("label", util.getText("context-resolve.error.label"));
    item.setAttribute("accesskey", util.getText("context-resolve.error.accesskey"));
    item.addEventListener("command", cmdShow, false);
  };

  req.findRedir(cmdData.uri); // BUG http://www.migre.me/contar-cliques/?url=7Kw0Y
}


/*
  var timeoutId = {};
  timeoutId.value = doc.defaultView.setTimeout(function() {
    configLongUrlFoundItems(mainMenu, cmdData, menuItem, copy, timeoutId);
  }, 0);
  menuItem.setAttribute("sublink-migreme-timeoutID", timeoutId.value);



  if (menuItem.getAttribute("sublink-migreme-timeoutID") !== timeoutId.value.toString()) {
    return;
  }
*/


function onPopupHidden(evt) {
  //console.log("onPopupHidden*");
  var popup = evt.originalTarget;
  popup.removeEventListener("popuphidden", onPopupHidden, false);
  evt.stopPropagation();
}



function updateRedirPath(menu, redirPath) {
  var doc = menu.ownerDocument;

  for (var idx = 0, len = redirPath.length; idx < len ; idx++) {
    var rid = "sublink-migreme-menu-resolvedurl-" + idx;
    if (doc.getElementById(rid) !== null) {
      continue;
    }

    if (idx > 0) {
      var prev = doc.getElementById("sublink-migreme-menu-resolvedurl-" + (idx - 1));
      prev.setAttribute("tooltiptext", redirPath[idx].delta + "\u202fms");
      if (prev.hasAttribute("disabled")) {
        prev.removeAttribute("image");
      } else {
        prev.setAttribute("image", "chrome://global/skin/icons/information-16.png");
      }

      var httpStatus = redirPath[idx].responseStatusCode;
      if (httpStatus > 0 && httpStatus !== 301) {
        var desc = util.getText("context-resolve.http-status.description", httpStatus);
        prev.setAttribute("description", desc);
      }
    }


    if (RedirServer.isRedirUri(redirPath[idx].longUri) === false) {
      return;
    }


    var currentRedir = menu.appendChild(doc.createElement("menuitem"));
    currentRedir.setAttribute("id", rid);

    var myUri = redirPath[idx].longUri; // TODO resolvedUri

    currentRedir.setAttribute("label", myUri.spec);
    currentRedir.classList.add("menuitem-iconic");
    currentRedir.setAttribute("image", "chrome://browser/skin/places/searching_16.png");

    var info = RedirServer.getInfoUrl(myUri);
    if (info === null) {
      currentRedir.setAttribute("disabled", "true");
      continue;
    }

    currentRedir.addEventListener("command", onRedirInfoCommand, false);
    currentRedir.addEventListener("click", onRedirInfoCommand, false);
    currentRedir.value = info;
    setStatusBar(currentRedir, info);
  }
}


function configMenuItemAction(menuItem, menuType) {
  switch (menuType) {
    case "popup-link":
      //menuItem.value = "copy";
      //menuItem.setAttribute("label", util.getText("context.link.loaded.label"));
      //menuItem.setAttribute("accesskey", util.getText("context.link.loaded.accesskey"));
      //break;

    case "popup-input-readonly":
      menuItem.value = "copy";
      menuItem.setAttribute("label", util.getText("context-resolve.copy.label"));
      menuItem.setAttribute("accesskey", util.getText("context-resolve.copy.accesskey"));
      break;

    case "popup-input-default":
      menuItem.value = "replace";
      menuItem.setAttribute("label", util.getText("context-resolve.replace.label"));
      menuItem.setAttribute("accesskey", util.getText("context-resolve.replace.accesskey"));
      break;

    default:
      console.trace();
      throw new Error(menuType);
  }
}


function configMenuItemRemember(menuItem) {
  menuItem.setAttribute("label", util.getText("context-resolve.remember.label"));
  menuItem.setAttribute("accesskey", util.getText("context-resolve.remember.accesskey"));
  menuItem.value = "add_library";
}



function configMenuItemConvert(menuItem, menuType) {
  switch (menuType) {
    case "popup-link":
      //menuItem.value = "convert_copy";
      //menuItem.setAttribute("label", util.getText("context.link.convertCopy.label", "migre.me"));
      //menuItem.setAttribute("accesskey", util.getText("context.link.convertCopy.accesskey"));
      //menuItem.setAttribute("image", "http://migre.me/favicon.ico");
      //menuItem.classList.add("menuitem-iconic");
      //break;

    case "popup-input-readonly":
      menuItem.value = "convert_copy";
      menuItem.setAttribute("label", util.getText("context-resolve.convert.copy.label", "migre.me"));
      menuItem.setAttribute("accesskey", util.getText("context-resolve.convert.copy.accesskey"));
      menuItem.setAttribute("image", "http://migre.me/favicon.ico");
      menuItem.classList.add("menuitem-iconic");
      break;

    case "popup-input-default":
      menuItem.value = "convert_replace";
      // TODO ver se tem salvo
      menuItem.setAttribute("label", util.getText("context-resolve.convert.replace.label", "migre.me"));
      menuItem.setAttribute("accesskey", util.getText("context-resolve.convert.replace.accesskey"));
      menuItem.setAttribute("image", "http://migre.me/favicon.ico");
      menuItem.classList.add("menuitem-iconic");
      break;

    default:
      console.trace();
      throw new Error(menuType);
  }
}


function finishResolvePopup(doc, cmdData, redirPath) {
  var menu = doc.getElementById("sublink-migreme-menu-resolve");
  if (menu === null) {
    // abort: menu is gone

    try {
      Cc["@mozilla.org/sound;1"].createInstance(Ci.nsISound).playEventSound(Ci.nsISound.EVENT_NEW_MAIL_RECEIVED);
    } catch (ex) {
      console.log(ex);
    }
    return;
  }

  if (menu.hasAttribute("disabled") === false) {
    // abort: simultaneous requests
    //return;
  }

  var info = redirPath[redirPath.length - 1];

  var favicons = Cc["@mozilla.org/browser/favicon-service;1"].getService(Ci.mozIAsyncFavicons);
  favicons.getFaviconURLForPage(info.longUri, {onComplete: function(favuri) {
      var urlFav = favuri ? favuri.spec
                          : "chrome://mozapps/skin/places/defaultFavicon.png";
      on2(doc, cmdData, redirPath, menu, info, urlFav);
  }});
}


function on2(doc, cmdData, redirPath, menu, info, fav) {

  menu.setAttribute("image", fav);

  menu.addEventListener("popuphidden", onPopupHidden, false); // stop propagation
  menu.setAttribute("label", info.longUri.prePath);
  switch (info.longUri.scheme) {
    case "http":
    case "https":
      menu.setAttribute("accesskey", "h");
      break;
    case "ftp":
      menu.setAttribute("accesskey", "f");
      break;
  }

  var cmdHandler = new OnLongUriCmds(cmdData,
                                     info.longUri,
                                     doc.defaultView.getBrowser().selectedBrowser);


  var pathMenuItem    = doc.getElementById("sublink-migreme-menu-resolve-tmp2");
  var copyMenuItem    = doc.getElementById("sublink-migreme-menu-resolve-tmp3");
  var convertMenuItem = doc.getElementById("sublink-migreme-menu-resolve-tmp4");
  var libraryItem     = doc.getElementById("sublink-migreme-menu-resolve-tmp5");

  //configMenuItemPath(pathMenuItem, fav, redirPath);
  var info = redirPath[redirPath.length - 1];
  var longUri = info.longUri;
  util.pageTitle(longUri, function(title) {
    configMenuItemPath(pathMenuItem, fav, redirPath, longUri, title);
  });


  pathMenuItem.removeAttribute("disabled");

  copyMenuItem.addEventListener("command", cmdHandler, false);
  copyMenuItem.removeAttribute("disabled");


  if (cmdData.uri.host === "migre.me") {//basedomain
    Components.utils.import("resource://sublink-migreme-4.1.0/redir-storage.js");
    var redirs = redirStorage.loadShortUrls(info.longUri);
    if (redirs.hasOwnProperty(cmdData.uri.spec)) {
      libraryItem.setAttribute("disabled", "true");
    } else {
      libraryItem.removeAttribute("disabled");
    }
    libraryItem.addEventListener("command", cmdHandler, false);

  } else {
    convertMenuItem.removeAttribute("disabled");
    convertMenuItem.addEventListener("command", cmdHandler, false);
  }
}


function findProperty(redirPath, name) {
  for (var idx = redirPath.length - 1; idx > -1; idx--) {
    var info = redirPath[idx];
    if (info.hasOwnProperty(name) && (info[name] !== null)) {
      return info[name];
    }
  }
  return null;
}

/*
function configMenuItemPath(menuItem, fav, redirPath) {
  var info = redirPath[redirPath.length - 1];
  var longUri = info.longUri;
  util.pageTitle(longUri, function(title) {
    configMenuItemPath2(menuItem, fav, redirPath, longUri, title);
  });
}*/


function configMenuItemPath(menuItem, fav, redirPath, longUri, title) {
  if (!title) {
    title = findProperty(redirPath, "htmlTitle");
  }

  var ct = findProperty(redirPath, "contentType");
  if (!title && ct) {
    if (ct.indexOf("image/") === 0) {
      var ct2 = ct.split("/", 2);
      title = util.getText("context-resolve.title.content-type.image.label", ct2[1].toUpperCase());
    } else if (ct.indexOf("text/") !== 0) {
      menuItem.setAttribute("description", ct);
    }
  }


  function onclick(evt) {
    util.openPage2(evt, longUri.spec, true);
  }


  if (!title) {
    // moz-anno:favicon:http://blog.mozilla.com/favicon.ico
    menuItem.setAttribute("image", fav);
    menuItem.classList.add("menuitem-iconic");
    menuItem.setAttribute("tooltiptext", longUri.spec);

  } else { // BUG se abrir o menu várias vezes, vai criar 2 linhas com title
    var titleItem = menuItem.parentNode.insertBefore(
                      menuItem.ownerDocument.createElement("menuitem"),
                      menuItem);

    titleItem.setAttribute("image", fav);
    titleItem.classList.add("menuitem-iconic");

    var ttt = title + "\n" + longUri.spec;
    titleItem.setAttribute("tooltiptext", ttt);
    menuItem.setAttribute("tooltiptext", ttt);

    titleItem.setAttribute("label", title);
    //titleItem.setAttribute("crop", "center");
    titleItem.setAttribute("default", "true");
    setStatusBar(titleItem, longUri.spec);
    //titleItem.setAttribute("disabled", "true");


    titleItem.addEventListener("command", onclick, false);
    titleItem.addEventListener("click", onclick, false);
  }


  menuItem.addEventListener("command", onclick, false);
  menuItem.addEventListener("click", onclick, false);

  menuItem.setAttribute("label", longUri.path);
  menuItem.setAttribute("crop", "center");
  menuItem.setAttribute("accesskey", "/");
  setStatusBar(menuItem, longUri.spec);


  /*
  if (!title) {
    menuItem.setAttribute("tooltiptext", longUri.spec);
    menuItem.setAttribute("label", longUri.path);
    menuItem.setAttribute("crop", "center");

  } else {
    menuItem.setAttribute("description", title);
    menuItem.setAttribute("tooltiptext", title + "\n" + longUri.spec);
    if (longUri.path.length > 40) {
      // ensure "description" will appear
      menuItem.setAttribute("label", longUri.path.substr(0, 36) + "\u2026");
    } else {
      menuItem.setAttribute("label", longUri.path);
    }
  }
  */


  // http://developer.yahoo.com/rss/#terms
  // http://api.search.yahoo.com/WebSearchService/rss/webSearch.xml?appid=yahoosearchwebrss&query=url%3Ahttp%3A%2F%2Fblogs.msdn.com%2Fbrada%2Farticles%2F371015.aspx&adult_ok=1

  // http://code.google.com/intl/pt-BR/apis/ajaxsearch/documentation/#fonje
  // http://ajax.googleapis.com/ajax/services/search/web?v=1.0&q=info:http%3A%2F%2Fblogs.msdn.com%2Fbrada%2Farticles%2F371015.aspx
}


/*
function hasRedirPathError(redirPath) {
  for (var idx = 0, len = redirPath.length - 1; idx < len ; idx++) {
    var httpStatus = redirPath[idx + 1].responseStatusCode;
    if (httpStatus === -200) {
      // redirectionLimit error
      return true;
    }
  }
  return false;
}
*/


function onRedirInfoCommand(evt) {
  var item = evt.target;
  var popup = item.parentNode;
  var items = popup.querySelectorAll("menuitem[image='chrome://global/skin/icons/information-16.png']");
  var close = items.length < 2;
  var url = item.value;
  util.openPage2(evt, url, close);
}



//copy / copy as / remember
function OnLongUriCmds(c, l, b) {
  this.cmdData = c;
  this.longUri = l;
  this.browserOrigin = b;
}


OnLongUriCmds.prototype = {

  handleEvent: function(evt) {
    var menuItem = evt.target;
    var cmd = menuItem.value;
    switch (cmd) {
      case "convert_copy":
      case "convert_replace":
      case "copy":
      case "replace":
      case "add_library":
        this[cmd]();
        break;
      default:
        console.trace();
        throw new Error(cmd);
    }
  },

  _isExtRedir: function(uri) {
    return uri.host === "migre.me";
  },

  add_library: function() {
    console.log("add_library", this.longUri.spec + "\n" + this.cmdData.toSource());
    if (this._isExtRedir(this.cmdData.uri) === false) {
      console.trace();
      throw new Error(this.cmdData.uri);
    }

    Components.utils.import("resource://sublink-migreme-4.1.0/redir-storage.js");
    redirStorage.saveShortUrl(this.longUri, this.cmdData.uri);
  },

  copy: function() {
    util.toClipboard(this.longUri.spec);
  },

  replace: function() {
    Components.utils.import("resource://sublink-migreme-4.1.0/run-input.js");
    PUBLIC_replaceCore(this.cmdData, this.longUri.spec, this.browserOrigin);
  },

  convert_copy: function() {
    var browser = this.browserOrigin;
    var chromeWnd = browser.ownerDocument.defaultView;

    function onSuccess(info) {
      util.toClipboard(info.shortUri.spec);
    }

    function onError(errData) {
      Components.utils.import("resource://sublink-migreme-4.1.0/err.js");
      notificationError(errData, browser);
    }

    function onComplete() {
      chromeWnd.setCursor("auto");
    }

    util.toClipboard("");
    chromeWnd.setCursor("progress");
    convertUrl(this.longUri, onSuccess, onError, onComplete);
  },


  convert_replace: function() {
    var data = this.cmdData;
    var browser = this.browserOrigin;
    var chromeWnd = browser.ownerDocument.defaultView;

    function onSuccess(info) {
      Components.utils.import("resource://sublink-migreme-4.1.0/run-input.js");
      PUBLIC_replaceCore(data, info.shortUri.spec, browser);
    }

    function onError(errData) {
      Components.utils.import("resource://sublink-migreme-4.1.0/err.js");
      notificationError(errData, browser);
    }

    function onComplete() {
      chromeWnd.setCursor("auto");
    }

    chromeWnd.setCursor("progress");
    convertUrl(this.longUri, onSuccess, onError, onComplete);
  }

}


function setStatusBar(item, text) {
  item.setAttribute("sublink-migreme-statusbartext", text);

  item.addEventListener("DOMMenuItemActive", function(evt) {
    setStatusBarCore(this, this.getAttribute("sublink-migreme-statusbartext"));
  }, false);

  item.addEventListener("DOMMenuItemInactive", function(evt) {
    setStatusBarCore(this, ""); // menuitem = this or evt.target
  }, false);
}

function setStatusBarCore(item, text) {
  item.ownerDocument.defaultView.XULBrowserWindow.setOverLink(text);
}


var m_cache = null;
function redirCache() {
  if (m_cache === null) {
    m_cache = new ObjCache(4);
  }
  return m_cache;
}


function ObjCache(len) {
  this._keys = new Array(len);
  this._values = new Array(len);
  this._nextIndex = 0;
}

ObjCache.prototype = {
  get: function(key) {
    var idx = this._keys.indexOf(key);
    if (idx === -1) {
      return null;
    }
    var val = this._values[idx];
    this._set(key, val);
    return val;
  },

  set: function(key, val) {
    var idx = this._keys.indexOf(key);

    if (idx > -1) {
      this._keys[idx] = null;
      this._values[idx] = null;
    }

    this._set(key, val);
  },

  _set: function(key, val) {
    var pos = this._nextIndex;
    this._keys[pos] = key;
    this._values[pos] = val;
    pos++;
    this._nextIndex = pos === this._keys.length ? 0 : pos;
  }

};
